/*
 * Copyright (c) 2016 CA.  All rights reserved.
 *
 * This software and all information contained therein is confidential
 * and proprietary and shall not be duplicated, used, disclosed or
 * disseminated in any way except as authorized by the applicable
 * license agreement, without the express written permission of CA.
 * All authorized reproductions must be marked with this language.
 *
 * EXCEPT AS SET FORTH IN THE APPLICABLE LICENSE AGREEMENT, TO THE
 * EXTENT PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS SOFTWARE
 * WITHOUT WARRANTY OF ANY KIND, INCLUDING WITHOUT LIMITATION, ANY
 * IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR
 * PURPOSE.  IN NO EVENT WILL CA BE LIABLE TO THE END USER OR ANY
 * THIRD PARTY FOR ANY LOSS OR DAMAGE, DIRECT OR INDIRECT, FROM THE
 * USE OF THIS SOFTWARE, INCLUDING WITHOUT LIMITATION, LOST PROFITS,
 * BUSINESS INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS
 * EXPRESSLY ADVISED OF SUCH LOSS OR DAMAGE.
 *
 */

package sample.restclient;

import sample.restservice.RequestType;
import sample.restservice.BasicService;
import java.io.BufferedWriter;
import java.io.IOException;

/**
 * This class acts as a client for the sample to make calls to the REST API
 *
 */
public class Client {

    private BasicService         service;
    private final String         baseUrl;

    //buffer for logging to a file
    private BufferedWriter       bw;

    //constants for logging purposes
    private static final int     SIZE     = 4;
    private static final int     REQ_HDR  = 0;
    private static final int     REQ_BODY = 1;
    private static final int     RES_HDR  = 2;
    private static final int     RES_BODY = 3;

    public Client(String authzHdrValue, String baseUrl, BufferedWriter outputWriter) {
        BasicService s = new BasicService(authzHdrValue);
        this.service = s;
        this.baseUrl = baseUrl;
        this.bw = outputWriter;
    }

    public static void print(String msg) {
       System.out.println(msg);
    }

    // a method for printing the request type that was made and a message
    private void printRequestURL(RequestType type, String msg){
        LogFile("Request:");
        LogFile("--");
        String request = type.toString() + " " + msg + " HTTP/1.1";
        LogFile(request);
    }

    // a method for pretty printing data fields of an SmObject
    public static String getPPrint(ObjectData data) throws IOException {
        if (data == null) return null;
        String result =  "";
        try {
           result = JSON.pp(data);
        } catch(IOException ex){
            print(ex.toString());
            throw ex;
        }
        return result;
    }

    // a method for pretty printing JSON data of an SmObject
    private String getPPrint(SmObject object) throws IOException {
        if (object == null) return null;
        String result = "";
        try {
           result = JSON.pp(object);
        } catch(IOException ex){
            print(ex.toString());
            throw ex;
        }
        return result;
    }

    // a method for pretty printing a list of objects
    private String getPPrint(LinkList object) throws IOException {
        if (object == null) return null;
        String result = "";
        try {
           result = JSON.pp(object);
        } catch(IOException ex){
            print(ex.toString());
            throw ex;
        }
        return result;
    }
    
    // a method for pretty printing a map of links
    private String getPPrint(LinkMap object) throws IOException {
        if (object == null) return null;
        String result = "";
        try {
           result = JSON.pp(object);
        } catch(IOException ex){
            print(ex.toString());
            throw ex;
        }
        return result;
    }


    // convert response string to object.  
    public SmObject JsonToSmObject(String response) throws IOException {
        if (response == null) return null;
        SmObject obj = null;
        try {
            obj = SmObject.fromJson(response);
            //pretty print response body
            LogFile(getPPrint(obj));
        } catch (Exception e) {
            //if pretty print fails, just log the response as is
            print(response);
            LogFile(response);
            bw.newLine();
            bw.newLine();
            return null;
        }
        bw.newLine();
        bw.newLine();
        return obj;
    } 

    // convert response string to object.  
    public LinkList JsonToLinkList(String response) throws IOException {
        if (response == null) return null;
        LinkList obj = null;
        try {
            obj = LinkList.fromJson(response);
            //pretty print response body
            LogFile(getPPrint(obj));
        } catch (Exception e) {
            //if pretty print fails, just log the response as is
            print(response);
            LogFile(response);
            bw.newLine();
            bw.newLine(); 
            return null;
        }
        bw.newLine();
        bw.newLine();
        return obj;
    } 

    // convert response string to object.  
    public LinkMap JsonToLinkMap(String response) throws IOException {
        if (response == null) return null;
        LinkMap obj = null;
        try {
            obj = LinkMap.fromJson(response);
            //pretty print response body
            LogFile(getPPrint(obj));
        } catch (Exception e) {
            //if pretty print fails, just log the response as is
            print(response);
            LogFile(response);
            bw.newLine();
            bw.newLine(); 
            return null;
        }
        bw.newLine();
        bw.newLine();
        return obj;
    } 


    // a method that wraps a RESTful GET request
    public SmObject getObject(String href) throws IOException {
      try {
        // string array to be populated by the BasicService
        String[] headers = new String[SIZE];
        //make HTTP GET request
        String response = service.processHTTPRequest(href, RequestType.GET, null, headers);

        //print request URL to log file
        printRequestURL(RequestType.GET, href);
        //print the rest of the HTTP data being sent and received
        if (headers != null) {
          LogFile(headers[REQ_HDR]);
          LogFile("");
          LogFile("Response:");
          LogFile("--");
          LogFile(headers[RES_HDR]);
        }
        return JsonToSmObject(response);
      } catch (IOException e) {
        print("Error while processing GET request");
        throw e;
      }
    }

    // a method that wraps a RESTful PUT request
    public SmObject putObject(SmObject obj) throws IOException {
      String href = obj.getSelf();
      try {
        //make HTTP PUT request
        String[] headers = new String[SIZE];
        ObjectData data = obj.getData();
        String response = service.processHTTPRequest(href, RequestType.PUT, data.toJson(), headers);

        //print request URL
        printRequestURL(RequestType.PUT, href);
        //print the rest of the HTTP data
        if (headers != null) {
          LogFile(headers[REQ_HDR]);

          ObjectData body = ObjectData.fromJson(headers[REQ_BODY]);
          LogFile(getPPrint(body));

          LogFile("");
          LogFile("Response:");
          LogFile("--");
          LogFile(headers[RES_HDR]);
        }
        return JsonToSmObject(response);
      } catch (IOException e) {
        print("Error while processing PUT request");
        throw e;
      }
    }

    // a method that wraps a RESTful POST request
    public SmObject postObjectData(SmObject parent, String xpsclass, ObjectData data, String hostinfo) throws IOException {
      String pref = (parent == null) ? baseUrl : parent.getSelf();
      String href = (parent == null) ? hostinfo + pref + '/' + xpsclass : pref + '/' + xpsclass;
     
      try {
        //make HTTP POST request
        String[] headers = new String[SIZE];
        String response = service.processHTTPRequest(href, RequestType.POST, data.toJson(), headers);

        //print HTTP request URL
        printRequestURL(RequestType.POST, href);
        //print the rest of the HTTP data
        if (headers != null) {
          LogFile(headers[REQ_HDR]);
          //print request data
          LogFile(getPPrint(data));

          LogFile("");
          LogFile("Response:");
          LogFile("--");
          LogFile(headers[RES_HDR]);
        }
        return JsonToSmObject(response);
      } catch (IOException e) {
        print("Error while processing POST request:");
        throw e;
      }
    }
    
    // a method that wraps a RESTful POST request
    public String postData(String postData, String hostinfo){

      String pref = baseUrl;
      String href = hostinfo + pref;
      String response = null;
        
      try {
        //make HTTP POST request
        String[] headers = new String[SIZE];
        response = service.processHTTPRequest(href, RequestType.POST, postData, headers);

        //print HTTP request URL
        printRequestURL(RequestType.POST, href);
        //print the rest of the HTTP data
        if (headers != null) {
            // log the request headers  
            LogFile(headers[REQ_HDR]);
            LogFile(headers[REQ_BODY]);

            // log the response headers
            LogFile("");
            LogFile("Response:");
            LogFile("--");
            LogFile(headers[RES_HDR]);
            LogFile(response);
            LogFile("");
        }
      } catch (Exception e) {
        print("Error while processing POST request:");
        e.printStackTrace();  
      }
      
      return response;
    }    

    // a method that wraps a RESTful DELETE request
    public void deleteObject(String href) throws Exception {
      try {
        //make HTTP request
        String[] headers = new String[SIZE];
        String response = service.processHTTPRequest(href, RequestType.DELETE, null, headers);

        //print HTTP request URL
        printRequestURL(RequestType.DELETE, href);
        //print the rest of the HTTP data
        if (headers != null) {
          LogFile(headers[REQ_HDR]);
          LogFile("");
          LogFile("Response:");
          LogFile("--");
          LogFile(headers[RES_HDR]);
        }

        JsonToSmObject(response);

      } catch (Exception e) {
        print("Error while processing DELETE request:");
        throw e;
      }
    }

    // a method for doing wildcard search; get links that are stored directly as a list
    public LinkList getLinkList(String href) throws IOException {
      try {
        //make GET request
        String[] headers = new String[SIZE];
        String response = service.processHTTPRequest(href, RequestType.GET, null, headers);

        //print request URL to log file
        printRequestURL(RequestType.GET, href);
        //print the rest of the HTTP data being sent and received
        if (headers != null) {
          LogFile(headers[REQ_HDR]);
          LogFile("");
          LogFile("Response:");
          LogFile("--");
          LogFile(headers[RES_HDR]);
        }
        return JsonToLinkList(response);
      } catch (IOException e) {
        e.printStackTrace();
        throw e;
      }
    }

    // a method for getting links that are stored as a map of lists
    public LinkMap getLinkMap(String href) throws IOException {
      try {
        //make GET request
        String[] headers = new String[SIZE];
        String response = service.processHTTPRequest(href, RequestType.GET, null, headers);
        //print request URL to log file
        printRequestURL(RequestType.GET, href);
        //print the rest of the HTTP data being sent and received
        if (headers != null) {
          LogFile(headers[REQ_HDR]);
          LogFile("");
          LogFile("Response:");
          LogFile("--");
          LogFile(headers[RES_HDR]);
        }
        
        return JsonToLinkMap(response);
        
      } catch (IOException e) {
        e.printStackTrace();
        throw e;
      }
    }

    //write output to a file
    private void LogFile(String str) {
      try {
        if (str == null) return;
        String s = str.replaceAll("\n", "\r\n");
        bw.write(s);
        bw.newLine();
      } catch (IOException ex) {
        ex.printStackTrace();
        return;
      }
    }

}
