/*
** Copyright (c) 2016 CA.  All rights reserved.
** This software may not be duplicated, disclosed or reproduced in whole or
** in part for any purpose except as authorized by the applicable license agreement,
** without the express written authorization of CA. All authorized reproductions
** must be marked with this language.
**
** TO THE EXTENT PERMITTED BY APPLICABLE LAW, CA PROVIDES THIS
** SOFTWARE “AS IS” WITHOUT WARRANTY OF ANY KIND, INCLUDING
** WITHOUT LIMITATION, ANY IMPLIED WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE OR NONINFRINGEMENT.  IN NO EVENT
** WILL CA BE LIABLE TO THE END USER OR ANY THIRD PARTY FOR ANY LOSS
** OR DAMAGE, DIRECT OR INDIRECT, FROM THE USE OF THIS MATERIAL,
** INCLUDING WITHOUT LIMITATION, LOST PROFITS, BUSINESS
** INTERRUPTION, GOODWILL, OR LOST DATA, EVEN IF CA IS EXPRESSLY
** ADVISED OF SUCH LOSS OR DAMAGE.
*/
/*======================================================================================
CA Single Sign-On RESTful API - Sample Client App (JavaScript)
========================================================================================
The purpose of this script is to demonstrate usage of the CA Single Sign-On RESTful API
========================================================================================
*/
/**
 * =====================================================================================
 * The first three functions in this block below:
 * getJSON(), deleteJSON(), and putJSON()
 * are the main functions for making a REST call for
 * either of CRUD operations using jQuery AJAX method.
 * =====================================================================================
*/
/**
 * @function getJSON
 * @desc Invokes ajax call for HTTP GET request
 * @param {string} hrefValue - links data of a returned JSON object.
 * @param {function} runFunction - either tree_process_list or tree_process_link
*/
function getJSON(hrefValue, runFunction) {
    return $.ajax({
        url: hrefValue,
        type: "GET",
        async: false,
        dataType: "json",
        headers: {
            "Authorization": "Bearer " + jwtToken.getValue()
        },
        success: runFunction,
        error: function (xhr) {
            errFunction(xhr.status, xhr.responseText);
        }
    });
}
/**
 * @function deleteJSON
 * @desc Invokes ajax call for HTTP DELETE request
 * @param {string} hrefValue - links data of a returned JSON object.
 * @param {function} runFunction
*/
function deleteJSON(hrefValue, runFunction) {
    $.ajax({
        url: hrefValue,
        method: "DELETE",
        headers: {
            "Authorization": "Bearer " + jwtToken.getValue()
        },
        success: runFunction,
        error: function (xhr) {
            errFunction(xhr.status, xhr.responseText);
        }
    });
}
/**
 * @function putJSON
 * @desc Invokes ajax call for HTTP PUT request (Update operation)
 * @param {string} hrefValue - links data of a returned JSON object.
 * @param {Object} obj - new object for data section
 * @param {function} runFunction
*/
function putJSON(hrefValue, obj, runFunction) {
    return $.ajax({
        url: hrefValue,
        method: "PUT",
        data: obj,
        headers: {
            "Authorization": "Bearer " + jwtToken.getValue()
        },
        success: runFunction,
        error: function (xhr) {
            errFunction(xhr.status, xhr.responseText);
        }
    });
}
/**
 * =====================================================================================
*/
/**
 * =====================================================================================
 * Five functions below are used for authentication (JWT Token)
 * =====================================================================================
*/
/**
 * @function CONFIG
 * @desc Stores address configuration
*/
var CONFIG = function () {
    var private = {
        'ORIGIN': document.location.origin,
        'SECURITYSERVICE': '/ca/api/sso/services/login/v1/token'
    };
    return {
        getValue: function (name) {
            return private[name];
        }
    };
}();
/**
 * @function getJwtToken
 * @desc Retrieves JWT Token
*/
function getJwtToken() {
    //  checks for JWT token
    if (jwtToken.getValue() == null) {
        //  fetches the token from the authentication service
        var uname = sessionStorage.getItem('uname');
        var passwd = sessionStorage.getItem('passwd');
        return $.ajax({
            url: CONFIG.getValue('ORIGIN') + CONFIG.getValue('SECURITYSERVICE'),
            type: "POST",
            async: false,
            dataType: "json",
            headers: {
                "Authorization": "Basic " + btoa(uname + ":" + passwd)
            },
            success: function (data) {
                jwtToken.setValue(data['sessionkey']);
            },
            error: function (xhr) {
                processLogout(xhr.status);
            }
        });
    }
}
/**
 * @function jwtToken
 * @desc Stores the JWT token value in sessionStorage
 * @param {string} newToken - new token value
 * @return {string}
*/
var jwtToken = window.jwtToken || {
        getValue: function () {
            return sessionStorage.getItem('sessionkey');
        },
        setValue: function (newToken) {
            sessionStorage.setItem('sessionkey', newToken);
            sessionStorage.removeItem('uname');
            sessionStorage.removeItem('passwd');
            return sessionStorage.getItem('sessionkey');
        },
        initialUrl: 'v1'
    };

/**
 * @function processLogout
 * @desc Removes token and redirects to the authentication page
*/
function processLogout(errStatus) {
    sessionStorage.clear();
    if(errStatus){
        window.location.href = "login.htm?errCode=" + errStatus;
    } else {
        window.location.href = 'login.htm';
    }    
}
/**
 * @function errFunction
 * @desc Handles error from CRUD operations
 * @param {integer} errStatus - Error code
*/
function errFunction(errStatus, errMessage) {
    if (errStatus == 401){
        processLogout(errStatus);
    } else if (errStatus != 500) {
        $('#json-renderer').html(errMessage);
    }
}
/**
 * =====================================================================================
*/
/**
 * @function prevHref
 * @desc Stores the URL of previously displayed object.
 * @param {string} previousURL
 * @return {string}
*/
var prevHref = function () {
    var val = '';
    return {
        getValue: function () {
            return val;
        },
        setValue: function (previousURL) {
            val = previousURL;
            return val;
        }
    };
}();
/**
 * @function viewMode
 * @desc Stores viewing mode (CRUD)
 * @param {string} newMode
 * @return {string} Current viewing mode
*/
var viewMode = function () {
    var crud = ['create', 'read', 'update', 'delete'];
    var currentMode = '';
    return {
        getValue: function () {
            return currentMode;
        },
        setValue: function (newMode) {
            if (newMode >= 0 && newMode <= 3) {
                currentMode = crud[newMode];
            }
            return currentMode;
        }
    };
}();
/**
 * @function notSupported
 * @desc Stores attributes that are not yet supported by the CA REST API for CRUD
 * @param {string} key
 * @return {integer}
*/
var notSupported = function () {
    var attributes = {
        "Definition": 1,
        "NestedVariablesLink": 1,
        "MetaData": 1,
        "IPAddresses": 1,
        "ValidateIdDifferentUser": 1,
        "ValidateIdentity": 1,
        "AuthorizationMappingLink": 1,
        "ValidationMappingLink": 1,
        "SessionAssuranceLink": 1,
        "SelfRegLink": 1,
        "DLPExcludedClassificationList": 1,
        "IsDLP": 1
    };
    return {
        getValue: function (key) {
            if (attributes.hasOwnProperty(key)) {
                return 1;
            }
            else {
                return 0;
            }
        }
    };
}();
/**
 * @function dataAttributes
 * @desc Stores all attributes of from a JSON edit object
 * @param {Object} attributesObj - The object retrieved from the attributes' section
 * @param {string} key - key variable
 * @param {string} key2 - key variable from a nested object
 * @return {Object|string|null}
*/
var dataAttributes = function () {
    var data = {};
    return {
        reset: function () {
            data = {};
            return data;
        },
        getKey2Attribute: function (key, key2) {
            if (data.hasOwnProperty(key)) {
                if (data[key].hasOwnProperty(key2)) {
                    return data[key][key2];
                }
                else {
                    return null;
                }
            }
            else {
                return null;
            }
        },
        setValue: function (attributesObj) {
            data = attributesObj;
            return data;
        }
    };
}();
/**
 * @function jsonObject
 * @desc Stores data a JSON object
 * @param {Object} attributesObj - The object retrieved from the attributes' section
 * @param {string} key - key variable
 * @param {string} key2 - key variable from a nested object
 * @return {Object|string|null}
*/
var jsonObject = function () {
    var data = {};
    return {
        getKey2Attribute: function (key, key2) {
            if (data.hasOwnProperty(key)) {
                if (data[key].hasOwnProperty(key2)) {
                    return data[key][key2];
                }
                else {
                    return null;
                }
            }
            else {
                return null;
            }
        },
        setValue: function (attributesObj) {
            data = attributesObj;
            return data;
        }
    };
}();
/**
 * @function newData
 * @desc Stores data for a new object
 * @param {string} key
 * @param {string} value
 * @return {Object}
*/
var newData = function () {
    var data = {};
    return {
        reset: function () {
            data = {};
            return data;
        },
        getValue: function () {
            return data;
        },
        setValue: function (key, value) {
            data[key] = value;
            return data;
        }
    };
}();
/**
 * @function createId
 * @desc Creates ID for nodes in the JStree
 * @param {string} keyValue
 * @return {string}
*/
var createId = function () {
    var counter = 0;
    return function (keyValue) { return keyValue + (++counter); };
}();
/**
 * @function formContents
 * @desc Stores the HTML code of a form used for displaying the attributes
 * @param {string} htmlElement - Code for an HTML element
 * @return {string} HTML code
*/
var formContents = function () {
    var formObjects = '';
    return {
        reset: function (htmlElement) { return formObjects = ''; },
        setObjects: function (htmlElement) { return formObjects += htmlElement; }
    };
}();
/**
 * @function pluralize_class_name
 * @desc Enhances the presentation of node names in JStree
 * @param {string} clz - class name
 * @return {string}
*/
function pluralize_class_name(clz) {
    return clz.bold().italics();
}
/**
 * @function tree_process_list
 * @desc Retrieves all children nodes of a selected object
 * @param {Object} node - part of callback function in the loadTree() function
 * @param {Object} data - stores links objects
 * @return {Objects}
*/
function tree_process_list(node, data) {
    var childNodes = [];
    var links = data.data;
    if (!links)
        return childNodes;
    var keys = Object.keys(links).sort();
    if (!keys)
        return childNodes;
    keys.forEach(function (className) {
        var c = {
            id: createId(className),
            text: pluralize_class_name(className),
            children: true,
            type: 'class',
            icon: 'jstree-file'
        };
        var children = links[className];
        c.children = children;
        children.forEach(function (ll) {
            var title = ll.path;
            /* Use only last portion of the path */
            ll.text = title.replace(/.+\//, '');
            ll.id = ll.href;
            ll.children = true;
            ll.type = 'link';
        });
        childNodes.push(c);
    });
    return childNodes;
}
/**
 * @function tree_process_link
 * @desc Retrieves the children link of an object
 * @param {Object} node - part of callback function in the loadTree() function
 * @param {Object} data - stores links objects
 * @return {Objects}
*/
function tree_process_link(node, data) {
    var items;
    var children = data.links.children;
    if (!children) {
        return [];
    }
    getJSON(children.href, function (list) {
        items = tree_process_list(node, list);
    });
    return items;
}
/**
 * @function booleanButton
 * @desc Supplies HTML code for radiobutton
 * @param {string} elementName
 * @param {string} value
 * @param {string} elementLabel
 * @param {string} c - status of selection (checked or '')
 * @return {string} HTML code
*/
function booleanButton(elementName, value, elementLabel, c) {
    return '<label class="radio-inline"><input type="radio" \
          name="' + elementName + '" value="' + value + '" ' + c + ' ' + enableInput(elementName)
        + ' onChange="updateData(name)" id="radio' + elementLabel + '"> \
          <span>' + elementLabel + '</span></label>';
}
/**
 * @function enableInput
 * @desc Determines status of input elements based on the current viewing mode
 * @param {string} key
 * @return {string}
*/
function enableInput(key) {
    var d = 'disabled';
    var value = dataAttributes.getKey2Attribute(key, 'value');
    if (viewMode.getValue() == 'update' && value != null) {
        d = '';
    }
    return d;
}
/**
 * @function updateData
 * @desc Updates data for a new object stored in newData.data
 * @param {string} key
*/
function updateData(key) {
    var checkboxData = [];
    var radiobuttonData = {};
    var attributeType = dataAttributes.getKey2Attribute(key, 'type');
    var handling = dataAttributes.getKey2Attribute(key, 'handling');
    var isList = dataAttributes.getKey2Attribute(key, 'isList');
    var inputs = $('#pb_' + key + ' input:checked');
    if (attributeType == 'String' && isList === false) {
        newData.setValue(key, $('#text_' + key).val());
    }
    else if (attributeType == 'Boolean') {
        newData.setValue(key, $.parseJSON($('input[name=' + key + ']:checked').val()));
    }
    else if (handling == 'Enumeration') {
        newData.setValue(key, $('input[name=' + key + ']:checked').val());
    }
    else if (attributeType == 'Link') {
        if (isList === true) {
            if (inputs.length > 0) {
                inputs.each(function () {
                    checkboxData.push({ 'href': $(this).val() });
                });
            }
            newData.setValue(key, checkboxData);
        }
        else {
            radiobuttonData['href'] = $('input[name=' + key + ']:checked').val();
            newData.setValue(key, radiobuttonData);
        }
    }
    else if (isList === true || handling == 'Flags') {
        if (inputs.length > 0) {
            inputs.each(function () {
                checkboxData.push($(this).val());
            });
        }
        newData.setValue(key, checkboxData);
    }
}
/**
 * @function inputType
 * @desc Creates HTML code representing an attribute from a JSON object
 * @param {string} elementType - type
 * @param {string} key
 * @param {string} value
 * @param {string} checkedStatus
 * @param {string} elementName
 * @return {string} HTML code
*/
function inputType(elementType, key, value, checkedStatus, elementName) {
    var spanText = '';
    var labelOpen = '';
    var labelClose = '';
    var booleanButtonStr = '';
    var tc = ''; //tc: true check
    var fc = ''; //fc: false check
    if (elementType != 'text') {
        labelOpen = '<label>';
        labelClose = '<span>' + key + '</span></label>';
    }
    if (elementType == 'checkbox') {
        value = key;
    }
    if (elementType == 'boolean') {
        if (value === true) {
            tc = 'checked';
        }
        else if (value === false) {
            fc = 'checked';
        }
        return booleanButton(key, 'true', 'Yes', tc) + booleanButton(key, 'false', 'No', fc);
    }
    else {
        return '<div class="' + elementType + '">' + labelOpen + '<input type="' + elementType + '" class="form-control" \
            id="' + elementType + '_' + elementName + '" name="' + elementName + '" value="' + value + '" ' + checkedStatus + ' '
            + enableInput(elementName) +
            ' onChange="updateData(name)">' +
            labelClose + '</div>';
    }
}
/**
 * @function enablePanel
 * @desc Disables/Enables input elements inside a panel body
 * @param {string} key
*/
function enablePanel(key) {
    if ($('#enable_' + key).is(':checked')) {
        $('#pb_' + key + ' :input').removeAttr('disabled');
        updateData(key);
    }
    else {
        $('#pb_' + key + ' :input').prop('disabled', true);
        newData.setValue(key, null);
    }
}
/**
 * @function panelContent
 * @desc Creates HTML code for a panel
 * @param {string} panelType
 * @param {string} key
 * @param {string} htmlBody
 * @param {string} useScroll
 * @param {string} value
 * @return {string} HTML code
*/
function panelContent(panelType, key, htmlBody, useScroll, value) {
    var checked = '';
    var disabled = 'disabled';
    if (value != null) {
        checked = 'checked';
    }
    if (viewMode.getValue() == 'update') {
        disabled = '';
    }
    return '<div class="panel panel-default ' + panelType + '"> \
          <div class="panel-heading"> \
          <h3 class="panel-title"><input type="checkbox" id="enable_' + key + '" \
          value="enable-check" ' + disabled + ' ' + checked + ' \
          onClick="enablePanel(&quot;' + key + '&quot;)"> ' + key + '</h3></div> \
          <div class="panel-body" id="pb_' + key + '" ' + useScroll + '>' + htmlBody + '</div></div>';
}
/**
 * @function clickLink
 * @desc Changes the focus of JStree upon click event on a link.
 * @param {string} id - node ID
*/
function clickLink(id) {
    $('#tree').jstree('activate_node', id);
}
/**
 * @function linkTableBody
 * @desc Creates HTML code for a row of a table used for displaying links
 * @param {string} key
 * @param {string} checkedStatus
 * @param {string} hrefValue
 * @param {string} pathValue
 * @param {string} descValue
 * @return {string} HTML code
*/
function linkTableBody(key, checkedStatus, hrefValue, pathValue, descValue) {
    var ot = ''; //ot: optionType
    if (dataAttributes.getKey2Attribute(key, 'isList') === true) {
        ot = 'checkbox';
    }
    else {
        ot = 'radio';
    }
    if (pathValue != '') {
        return '<tr><td><input id="' + ot + '_' + pathValue + '" type="' + ot + '" ' + checkedStatus + ' '
            + enableInput(key) + ' name="' + key + '" value="' + hrefValue + '"\
            onChange="updateData(&quot;' + key + '&quot;)"></td>\
            <td><a id="#' + hrefValue + '" href="#" onClick="clickLink(id)" class="title-link">'
            + pathValue + '</a></td><td>' + descValue + '</td></tr>';
    }
    else {
        return '';
    }
}
/**
 * @function linkTable
 * @desc Creates HTML code for a table for displaying links
 * @param {string} htmlBody - HTML code for table body
 * @return {string} HTML code
*/
function linkTable(htmlBody) {
    if (htmlBody != '') {
        return '<table class="table" style="font-size:13"> \
              <thead><tr><th></th><th>path</th><th>desc</th></tr></thead> \
              <tbody>' + htmlBody + '</tbody> \
            </table>';
    }
    else {
        return '';
    }
}
/**
 * @function getLinkProperty
 * @desc Retrieves value of a link's property
 * @param {Object} valueObj - current object
 * @param {string} property (i.e. href, path, or desc)
 * @return {string}
*/
function getLinkProperty(valueObj, property) {
    if (valueObj.hasOwnProperty(property)) {
        return valueObj[property];
    }
    else {
        return '';
    }
}
/**
 * @function getSelect
 * @desc Retrieves object available to an attribute
 * @param {Object} x - a value's object
 * @return {Object} objHandling - list of links
*/
function getSelect(x) {
    var objHandling;
    if (x.hasOwnProperty('href')) {
        getJSON(x.href, function (data) {
            objHandling = data.data;
        });
    }
    return objHandling;
}
/**
 * @function getAvailableLinks
 * @desc Retrieves all links available to an attribute
 * @param {string} key
 * @param {Object} links - links' object
 * @param {Object} sl - selected links' object
 * @return {Object} linkrows - HTML code for links' table
*/
function getAvailableLinks(key, links, sl) {
    var linkrows = '';
    var checked = '';
    $.each(links, function (key2, v2) {
        if (typeof v2 == 'object') {
            if (v2.hasOwnProperty('href')) {
                if (sl[v2.href] === true) {
                    checked = 'checked';
                }
                else {
                    checked = '';
                }
                if (viewMode.getValue() == 'update') {
                    linkrows += linkTableBody(key, checked, v2.href, v2.path, v2.desc);
                }
                else {
                    if (checked == 'checked') {
                        linkrows += linkTableBody(key, checked, v2.href, v2.path, v2.desc);
                    }
                }
            }
        }
    });
    return linkrows;
}
/**
 * @function getSelectedLinks
 * @desc Retrieves links selected for an attribute
 * @param {Object} x - a value's object
 * @return {Object} newObj - with reversed key:value -> value:key
*/
function getSelectedLinks(data) {
    var hrefValue = '';
    var newObj = {};
    if (data.value != null) {
        if (data.isList === true) {
            $.each(data.value, function (key2, v2) {
                hrefValue = getLinkProperty(v2, 'href');
                newObj[hrefValue] = true;
            });
        }
        else {
            hrefValue = getLinkProperty(data.value, 'href');
            newObj[hrefValue] = true;
        }
    }
    return newObj;
}
/**
 * @function getScroll
 * @desc Adds scroll to a specific attribute with length of data > 10
 * @param {Object} data - a value's object
 * @return {string} CSS style
*/
function getScroll(data) {
    var style = '';
    if (data instanceof Array && data.length > 10) {
        style = 'style="height:200px; overflow:auto"';
    }
    else if (data != null) {
        if (Object.keys(data).length > 10) {
            style = 'style="height:200px; overflow:auto"';
        }
    }
    return style;
}
/**
 * @function checkProperty
 * @desc Retrieves value of an object's property
 * @param {Object} valueObj - current object
 * @param {string} property (i.e. href, path, or desc)
 * @return {string}
*/
function checkProperty(valueObj, property) {
    if (valueObj.hasOwnProperty(property)) {
        return valueObj[property];
    }
    else {
        return '';
    }
}
/**
 * @function buildAttributeView
 * @desc Creates a complete HTML element for an attribute
 * @param {string} key
 * @param {string} value
 * @return {string} HTML code
*/
function buildAttributeView(key, value) {
    /* str is a final HTML string of an attribute. */
    var str = '';
    /* subStr is a sub string used for building a final HTML
    string for an attribute. */
    var subStr = '';
    /* attributeType is used for storing the type of HTML element of an attribute. */
    var attributeType = '';
    /* checked is a string used to mark a checkbox or radiobutton as selected. */
    var checked = '';
    /* newObj is used for storing a new key-value pair object converted from
    an array. The elements of this object are values selected from a list of
    available options. */
    var newObj = {};
    var linksObj = {};
    /* hrefValue, pathValue, and desc are used for storing properties' values of a link */
    var hrefValue, pathValue, descValue;
    var needScroll = '';
    if (typeof value == 'object') {
        /* Check if the type of attribute is a String or Link. */
        if (value.type == 'String') {
            if (typeof value.value == 'object' && value.value != null) {
                /* Some attributes might have KeyValue as the handling value,
                for instance sm.agent-configs. KeyValue means that its key and value are
                concatenated as a string. The delimiter used here is the encryption mode bit
                enclosed by equal signs (i.e. '=0=' or '=1='). '=0=' means the value is
                not encrypted whereas '=1=' means it is encrypted.
                The following section below checks for such a condition and remove '=0'
                in order to improve readability of the key and value information. */
                if (value.handling == 'KeyValue') {
                    $.each(value.value, function (key2, v2) {
                        subStr += inputType('text', '', v2.replace('=0', ''), '', key);
                    });
                    needScroll = getScroll(value.value);
                }
                else {
                    $.each(value.value, function (key2, v2) {
                        subStr += inputType('text', '', v2, '', key);
                    });
                }
            }
            else {
                subStr += inputType('text', key, value.value, '', key);
            }
        }
        else if (value.type == 'Link') {
            if (viewMode.getValue() == 'update') {
                //check for available options
                if (value.hasOwnProperty('select')) {
                    if (typeof value.select == 'object') {
                        linksObj = getSelect(value.select);
                        subStr = linkTable(getAvailableLinks(key, linksObj, getSelectedLinks(value)));
                        if (viewMode.getValue() == 'update') {
                            needScroll = getScroll(linksObj);
                        }
                        else {
                            needScroll = getScroll(getSelectedLinks(value));
                        }
                    }
                }
            }
            else {
                if (value.value instanceof Array) {
                    var linkrows = '';
                    $.each(value.value, function (key2, v2) {
                        hrefValue = checkProperty(v2, 'href');
                        pathValue = checkProperty(v2, 'path');
                        descValue = checkProperty(v2, 'desc');
                        linkrows += linkTableBody(key, 'checked', hrefValue, pathValue, descValue);
                    });
                    subStr = linkTable(linkrows);
                }
                else {
                    hrefValue = checkProperty(value, 'href');
                    pathValue = checkProperty(value, 'path');
                    descValue = checkProperty(value, 'desc');
                    subStr = linkTable(linkTableBody(key, 'checked', hrefValue, pathValue, descValue));
                }
            }
        }
        else {
            /* If the type of attribute is not String or Link then check if it is
            Boolean or Integer. */
            if (value.type == 'Boolean') {
                attributeType = 'boolean';
            }
            else if (value.type == 'Integer') {
                if (value.handling == 'Flags') {
                    attributeType = 'checkbox';
                }
                else if (value.handling == 'Enumeration') {
                    attributeType = 'radio';
                }
            }
            /* Check for available options. */
            if (value.hasOwnProperty('bits')) {
                /* Convert array object to a key-value object. */
                if (value.value != null) {
                    if (typeof value.value != 'object') {
                        newObj[value.value] = true;
                    }
                    else {
                        $.each(value.value, function (i, val) {
                            newObj[val] = true;
                        });
                    }
                }
                /* Use a new key-value object (i.e. newObj) for marking
                the selected options when displaying all of available options */
                $.each(value.bits, function (key2, v2) {
                    if (newObj.hasOwnProperty(key2)) {
                        checked = 'checked';
                    }
                    else {
                        checked = '';
                    }
                    subStr += inputType(attributeType, key2, key2, checked, key);
                });
                needScroll = getScroll(value.bits);
            }
            else if (attributeType != '' && value.value != null) {
                subStr += inputType(attributeType, key, value.value, checked, key);
            }
        }
    }
    str = '<div class="form-group">' + subStr + '</div>';
    return panelContent('panel-info', key, str, needScroll, value.value);
}
/**
 * @function getEdit
 * @desc Retrives a link to JSON edit object
 * @param {Object} x - a JSON object
 * @return {Object} objHandling
*/
function getEdit(x) {
    var objHandling;
    getJSON(x.links.editinfo.href, function (data) {
        objHandling = data.attributes;
    });
    dataAttributes.setValue(objHandling);
    return objHandling;
}
/**
 * @function iterate
 * @desc Iterates through attributes inside a JSON object
 * @param {Object} data - a JSON object
*/
function iterate(data) {
    /* jQuery function to iterate through objects or array-like objects.
    Data from JSON object is in key-value pair format.*/
    $.each(data, function (key, value) {
        if (notSupported.getValue(key) == 0) {
            newData.setValue(key, value.value);
            formContents.setObjects(buildAttributeView(key, value));
        }
    });
}
/**
 * @function clickButton
 * @desc Executes CRUD operations
 * @param {string} id - JStree node's ID
 * @param {string} m - current viewing mode
*/
function clickButton(id, m) {
    var tree = $('#tree').jstree(true);
    viewMode.setValue(m);
    switch (m) {
        case 0:
            $('#tree').jstree('activate_node', id);
            break;
        case 1:
            console.log(JSON.stringify(newData.getValue()));
            putJSON(id, JSON.stringify(newData.getValue()), function (data) { });
            $('#tree').jstree('activate_node', id);
            break;
        case 2:
            $('#tree').jstree('activate_node', id);
            break;
        case 3:
            deleteJSON(id, function (data) { });
            tree.delete_node(id);
            clearForm();
            break;
    }
}
/**
 * @function continueCreate
 * @desc Loads template and default value for a new object
*/
function continueCreate() {
    var buttons = '';
    var bodyString = '';
    var el = ''; //el: editLink
    var sl = ''; //sl: selfLink
    var nh = $('input[name="new"]:checked').val(); //nh: newHref
    var url = window.location.href;
    viewMode.setValue(2);
    clearForm();
    newData.reset();
    getJSON(nh, function (data) {
        el = data.links.editinfo.href;
        sl = data.links.self.href;
    });
    getJSON(el, function (data) {
        dataAttributes.setValue(data.attributes);
        buttons = getButtons(sl);
        formContents.setObjects('<form id="myform">' + buttons);
        iterate(data['attributes']);
        formContents.setObjects('</form>');
        bodyString = panelContent('panel-primary', 'attributes', formContents.setObjects(''), '', '');
        //display the form for a new object
        w2ui['layout'].content('main', bodyString);
    });
}
/**
 * @function getChildrenOptions
 * @desc Retrieves options available for creating a new child object
 * @param {string} hrefValue
*/
function getChildrenOptions(hrefValue) {
    var str = '';
    var bodyString = '';
    var hasChildren = false;
    getJSON(hrefValue + '/children', function (data) {
        if (data.hasOwnProperty('links')) {
            str = getButtonContinueCancel(hrefValue);
            $.each(data.links['new'], function (key, value) {
                hasChildren = true;
                str += '<input id="radio' + key + '" type="radio" name="new" value="' + value + '"><span> ' + key + '</span><br>';
            });
            bodyString = panelContent('panel-primary', 'Options available for new object', str, '', '');
            //displays the options
            w2ui['layout'].content('main', bodyString);
        }
    });
    if (hasChildren === false) {
        alert("This object does not have options for children! ");
    }
}
/**
 * @function htmlButton
 * @desc Creates buttons for CRUD
 * @param {string} hrefValue
 * @param {integer} m - current viewing mode
 * @return {string} HTML code
*/
function htmlButton(hrefValue, m) {
    var mode = ['New', 'Save', 'Edit', 'Remove', 'Continue', 'Cancel'];
    var subMenu = '';
    var aStyle = ''; //additionalStyle
    var onClick = 'onClick="clickButton(&quot;' + hrefValue + '&quot;, ' + m + ')"';
    var glyphicon = {
        "New": "glyphicon-plus",
        "Edit": "glyphicon-pencil",
        "Save": "glyphicon-floppy-disk",
        "Remove": "glyphicon-trash",
        "Continue": "glyphicon-ok",
        "Cancel": "glyphicon-remove"
    };
    if (m == 0) {
        onClick = 'onClick="getChildrenOptions(&quot;' + hrefValue + '&quot;)"';
    }
    else if (m == 4) {
        onClick = 'onClick="continueCreate()"';
    }
    return '<a ' + aStyle + ' id="button' + mode[m] +
        '" href="#" ' + onClick + '\
          class="title-link"><span class="glyphicon ' +
        glyphicon[mode[m]] + '"></span> ' + mode[m] + '\
          </a>' + subMenu;
}
/**
 * @function getButtonContinueCancel
 * @desc Provides buttons for Create operation
 * @param {string} hrefValue
 * @return {string} HTML code
*/
function getButtonContinueCancel(hrefValue) {
    var button = '<ul class="nav nav-pills nav-justified">';
    for (var i = 4; i <= 5; i++) {
        button += '<li role="presentation">' + htmlButton(hrefValue, i) + '</li>';
    }
    button += '</ul>';
    return button;
}
/**
 * @function getButtons
 * @desc Wraps CRUD buttons for presentation according CAStrap CSS
 * @param {string} hrefValue
 * @return {string} HTML code
*/
function getButtons(hrefValue) {
    var buttons = '<ul class="nav nav-pills nav-justified">';
    for (var i = 0; i <= 3; i++) {
        buttons += '<li role="presentation">' + htmlButton(hrefValue, i) + '</li>';
    }
    buttons += '</ul>';
    return buttons;
}
/**
 * @function loadTree
 * @desc Loads JStree
*/
function loadTree() {
    $('#tree').jstree({
        'core': {
            'animation': 0,
            'check_callback': true,
            'data': function (node, cb) {
                var url; // holds the url value
                var runFunction; // determines which tree function to run
                if (node.id === '#') {
                    url = jwtToken.initialUrl;
                    runFunction = tree_process_list;
                }
                else {
                    url = node.id;
                    runFunction = tree_process_link;
                }

                // Get data from the RESTful API server
                getJSON(url, function (data) {
                    var items = runFunction(node, data);
                    cb.call(this, items);
                });
            } // data
        },
    }); //#tree
}
/**
 * @function clearForm
 * @desc Clears the content of the HTML form
*/
function clearForm() {
    var g = w2ui['layout'].content('main');
    if (g) {
        w2ui['layout'].content('main', '');
        formContents.reset('');
        $('#json-renderer').html("");
    }
}

/**
 * @function modifier
 * @desc Modifies links displayed in JSON format
 * @param {string} key
 * @param {string} value
 * @return {string} value/modified value
*/
function modifier(key, value) {
  // Filtering out properties
  if (key == 'href') {
    return "<a href=# onclick=clickLinkInJSON('"+ value + "')>" + value + "</a>";
  }
  return value;
}

/**
 * @function loadRawJSON
 * @desc Loads raw JSON from a given URL
 * @param {string} url
*/
function loadRawJSON(url){
    if(url.length > 0){
        $('#url').val(url);
        getJSON(url, function (obj){
            $('#json-renderer').html(JSON.stringify(obj, modifier, 2));
        });
    }
}

/**
 * @function loadJSONFromInputURL
 * @desc Invokes loadRawJSON() from event method keypress in a textbox
*/
function loadJSONFromInputURL(){
    loadRawJSON($('#url').val());
    urlStack1.pushValue($('#url').val());
}

/**
 * @function clickLinkInJSON
 * @desc Loads JSON data when a link is clicked
 * @param {string} url
*/
function clickLinkInJSON(url){
	$('#url').val(url);
	loadJSONFromInputURL();
}

/**
 * @function urlStack1
 * @desc Stores history for the Back button
 * @return {string|array} url from the stack
*/
var urlStack1 = function () {
    var data = new Array();
    return {
        reset: function () {
            data = [];
        },
        getValue: function (){
            return data;
        },
        popValue: function () {  
            if(data.length > 0){
                var tempURL = data.pop();
                urlStack2.pushValue(tempURL);
                return tempURL;
            } else {
                return "";
            }            
        },
        pushValue: function (value) {
			data.push(value);
        }
    };
}();

/**
 * @function urlStack2
 * @desc Stores history for the Forward button
 * @return {string|array} url from the stack
*/
var urlStack2 = function () {
    var data = new Array();
    return {
        reset: function () {
            data = [];
        },
        getValue: function (){
            return data;
        },
        popValue: function () {  
            if(data.length > 0){
                var tempURL = data.pop();
                urlStack1.pushValue(tempURL);
                return tempURL;
            } else {
                return "";
            }            
        },
        pushValue: function (value) {
			data.push(value);
        }
    };
}();

/**
 * @function backButtonClick
 * @desc Goes back in the url's history
*/
function backButtonClick(){  
    if(urlStack2.getValue().length == 0){
        urlStack1.popValue();
    }  
    var prevURL = urlStack1.popValue();    
    loadRawJSON(prevURL);
}

/**
 * @function forwardButtonClick
 * @desc Goes forward in the urls' history
*/
function forwardButtonClick(){
    if(urlStack1.getValue().length == 0){
        urlStack2.popValue();
    }
    var prevURL = urlStack2.popValue();
    loadRawJSON(prevURL);
}

/* MAIN FUNCTION */
$(function () {
    var pstyle = 'background-color: #F5F6F7; border: 1px solid #dfdfdf; padding: 5px;';
    var header = '<div class="global-header">\
                  <table width="100%""><tr><td>\
                  <div class="title" style="top:5px"> \
                    <p id="product-name" style="line-height:3px;">RESTful API - Policy Store</p> \
                    <p id="feature-name" style="line-height:25px;">Sample App</p> \
                  </div>\
                  </td><td align="right" style="padding-right: 10px;">\
                  <a href="#" onClick="processLogout()"><h3>Logout</h3></a>\
                  </td></tr></table>\
                 </div>';
    var urlBox = '<div class="input-group"><span class="input-group-btn">\
                 <button type="button" class="btn btn-primary btn-sm" onclick=backButtonClick()><</button>\
                 <button type="button" class="btn btn-primary btn-sm" onclick=forwardButtonClick()>></button></span>\
                 <input type="text" class="form-control" id="url"  onkeypress=if(event.keyCode==13)loadJSONFromInputURL()>\
                 </div>';
    var buttons = '';
    var bodyString;
    getJwtToken();
	
    /* Setting up the layout */
    $('body').w2layout({
        name: 'layout',
        panels: [
            /* The top panel is where the information about this sample app is located */
            { type: 'top', size: 60, content: header },
            /* The left panel is for the JSON tree */
            { type: 'left', size: '20%', style: pstyle, resizable: true, content: '<div id="tree"/>' },
            /* The main panel is in the middle of the screen
            where the data will be displayed in an HTML form format */
            { type: 'main', style: pstyle, resizable: true },
            /* The right panel displays data in the JSON format of a selected object */
            { type: 'right', size: '40%', style: pstyle, resizable: true, content: urlBox + '<pre style="margin-left: 0.4em;margin-top: 5px;" id="json-renderer"></pre>' },
        ]
    });
    /* Load the tree */
    loadTree();
	
    /* This section of code will be executed when the selected node on the tree
    is changed. It will recreate the HTML form containing the appropriate
    objects according to the returned JSON object. */
    $('#tree').on("changed.jstree", function (e, data) {
        var node = data.instance.get_node(data.selected[0]).original;
        /* Remove the form in the middle pane when the selected object is changed */
        clearForm();
        if (node != null) {
            if (prevHref.getValue() != node.href) {
                viewMode.setValue(1);
                newData.reset();
            }
            if ('link' === node.type) {
                buttons = getButtons(node.href);
                prevHref.setValue(node.href);
                $('#url').val(node.href);
                urlStack1.pushValue(node.href);                
                getJSON(node.href, function (obj) {
                    /* Stores object to jsonObject.data */
                    jsonObject.setValue(obj);
					
                    /* Render the raw JSON data in the the right pane of the screen */
                    $('#json-renderer').html(JSON.stringify(obj, modifier, 2));
					
                    /* The ID for the HTML form is static and therefore required
                    to be set before dynamically generating the content of the form. */
                    formContents.setObjects('<form id="myform">' + buttons);
					
                    /* Call the function to iterate through data content of
                    the returned JSON object and generate HTML object accordingly.
                    getEdit() returns a JSON object prepared for edit purpose.
                    The edit object has more information including how an attribute
                    should be handled and also all of options available to that
                    particular attribute. */
                    iterate(getEdit(obj));
					
                    /* Close the tag of the HTML form */
                    formContents.setObjects('</form>');
					
                    /* Render the form container object data in the middle pane of the screen.
                    formContents.setObjects('') will also return the latest string value
                    that have been concatenated earlier. */
                    bodyString = panelContent('panel-primary', 'attributes', formContents.setObjects(''), '', '');
                    w2ui['layout'].content('main', bodyString);
                });
            }
        }
    });
});
